# -*- coding: utf-8 -*-

#
# Copyright 2012, 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import json
import csv
import sys


from datetime import datetime
from .world import world, res_filename, logged_wait
from nose.tools import eq_, assert_less

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED
from bigml.api import FAULTY
from bigml.api import UPLOADING
from bigml.api import get_status


from . import read_source_steps as read

#@step(r'I create a data source uploading a "(.*)" file$')
def i_upload_a_file(step, file):
    resource = world.api.create_source(res_filename(file), \
        {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])

#@step(r'I create a data source uploading a "(.*)" file using a project$')
def i_upload_a_file_with_project_conn(step, file):
    resource = world.api.create_source(res_filename(file))
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])

#@step(r'I create a data source from stdin uploading a "(.*)" file$')
def i_upload_a_file_from_stdin(step, file):
    file_name = res_filename(file)
    with open(file_name, 'rb') as file_handler:
        resource = world.api.create_source(file_handler, \
            {'project': world.project_id})
        # update status
        world.status = resource['code']
        world.location = resource['location']
        world.source = resource['object']
        # save reference
        world.sources.append(resource['resource'])


#@step(r'I create a data source uploading a "(.*)" file with args "(.*)"$')
def i_upload_a_file_with_args(step, file, args):
    args = json.loads(args)
    args.update({'project': world.project_id})
    resource = world.api.create_source(res_filename(file), args)
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])

#@step(r'I create a data source using the url "(.*)"')
def i_create_using_url(step, url):
    resource = world.api.create_source(url, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])

#@step(r'I create a data source using the connection ".*"')
def i_create_using_connector(step, connector):
    resource = world.api.create_source(connector, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])

#@step(r'I create a data source from inline data slurped from "(.*)"')
def i_create_using_dict_data(step, data):
    # slurp CSV file to local variable
    mode = 'rb'
    if sys.version > '3':
        mode = 'rt'
    with open(res_filename(data), mode) as fid:
        reader = csv.DictReader(fid)
        dict_data = [row for row in reader]
    # create source
    resource = world.api.create_source(dict_data,
                                       {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])

#@step(r'I create a data source uploading a "(.*)" file in asynchronous mode$')
def i_upload_a_file_async(step, file):
    resource = world.api.create_source(res_filename(file),
                                       {'project': world.project_id},
                                       async_load=True)
    world.resource = resource

#@step(r'I wait until the source has been created less than (\d+) secs')
def the_source_has_been_created_async(step, secs):
    start = datetime.utcnow()
    read.i_get_the_source(step, world.source['resource'])
    status = get_status(world.resource)
    count = 0
    delta = int(secs) * world.delta
    while status['code'] == UPLOADING:
        count += 1
        logged_wait(start, delta, count, "source")
        read.i_get_the_source(step, world.source['resource'])
        status = get_status(world.source)
    eq_(world.resource['code'], HTTP_CREATED)
    # update status
    world.status = world.resource['code']
    world.location = world.resource['location']
    world.source = world.resource['object']
    # save reference
    world.sources.append(world.resource['resource'])

#@step(r'I wait until the source status code is either (\d) or (\d) less than (\d+)')
def wait_until_source_status_code_is(step, code1, code2, secs):
    start = datetime.utcnow()
    delta = int(secs) * world.delta
    read.i_get_the_source(step, world.source['resource'])
    status = get_status(world.source)
    count = 0
    while (status['code'] != int(code1) and
           status['code'] != int(code2)):
        count += 1
        logged_wait(start, delta, count, "source")
        read.i_get_the_source(step, world.source['resource'])
        status = get_status(world.source)
    eq_(status['code'], int(code1))

#@step(r'I wait until the source is ready less than (\d+)')
def the_source_is_finished(step, secs):
    wait_until_source_status_code_is(step, FINISHED, FAULTY, secs)

#@step(r'I update the source with params "(.*)"')
def i_update_source_with(step, data="{}"):
    resource = world.api.update_source(world.source.get('resource'), json.loads(data))
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)

#@step(r'the source exists and has args "(.*)"')
def source_has_args(step, args="{}"):
    args = json.loads(args)
    for key, value in list(args.items()):
        if key in world.source:
            eq_(world.source[key], value,
                "Expected key %s: %s. Found %s" % (key, value, world.source[key]))
        else:
            assert False, "No key %s in source." % key



# -*- coding: utf-8 -*-
#
# Copyright 2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
A fast building local Predictive Anomaly Detector.
This module defines an Anomaly Detector to score anomalies in a dataset locally
or embedded into your application without needing to send requests to
BigML.io.
The module is also designed for situations when it is desirable to be able to
build the anomaly detector very quickly from an external representation.
It also offers the ability to load its contents from a cache system like
Redis or memcache. The `get` method of the cache system has to be passed
in the `cache_get` argument and the hash for the storage should be the
corresponding anomaly ID.
Example usage (assuming that you have previously set up the BIGML_USERNAME
and BIGML_API_KEY environment variables and that you own the model/id below):
from bigml.api import BigML
from anomaly import Anomaly
import redis
r = redis.Redis()
# First build as you would any core Anomaly object:
anomaly = Anomaly('anomaly/5126965515526876630001b2')
# Store a serialized version in Redis
anomaly.dump(cache_set=r.set)
# (retrieve the external rep from its convenient place)
# Speedy Build from external rep
anomaly = Anomaly('anomaly/5126965515526876630001b2', cache_get=r.get)
# Get scores same as always:
anomaly.anomaly_score({"src_bytes": 350})
"""


import math
import msgpack
import json

from bigml.predicate_utils.utils import OPERATOR_CODE, PREDICATE_INFO_LENGTH
from bigml.predicate_utils.utils import apply_predicates
from bigml.api import FINISHED
from bigml.api import get_status, get_api_connection, get_anomaly_id
from bigml.basemodel import get_resource_dict
from bigml.modelfields import ModelFields
from bigml.util import cast


DEPTH_FACTOR = 0.5772156649
PREDICATES_OFFSET = 3

def build_tree(children, node=None):
    """Builds a compressed version of the tree structure as an list of
    lists. Starting from the root node, that is represented by a list:
        [len(children), children1, children2, etc.]
    And each child is represented by a list whose elements are:
        [weight, len(predicates), operator_code, field, value, term, missing,
         ..., len(children), children_nodes_list*]
    """
    outer = node if node else list()
    outer.append(len(children))
    for child in children:
        inner = list()
        inner.append(child.get('weight'))
        predicates = child.get('predicates')
        if predicates and not (predicates is True or predicates == [True]):
            predicates = [x for x in predicates if x is not True]
            inner.append(len(predicates))
            for pred in predicates:
                operation = pred.get('op')
                value = pred.get('value')
                missing = False
                if operation.endswith("*"):
                    operation = operation[0: -1]
                    missing = True
                elif operation == 'in' and None in value:
                    missing = True

                inner.append(OPERATOR_CODE.get(operation))
                inner.append(pred.get('field'))
                inner.append(value)
                inner.append(pred.get('term'))
                inner.append(missing)
        else:
            inner.append(0)
        next_gen = child.get('children')
        if next_gen:
            build_tree(next_gen, node=inner)
        else:
            inner.append(0)
        outer.append(inner)

    return outer


def calculate_depth(node, input_data, fields, depth=0):
    """Computes the depth in the tree for the input data

    """

    weight = node[0] if node[0] else 1
    num_predicates = node[1]
    num_children = node[2 + (5 * num_predicates)]

    if not apply_predicates(node, input_data, fields):
        return depth

    depth += weight
    if num_children > 0:
        start = PREDICATES_OFFSET + (PREDICATE_INFO_LENGTH * num_predicates)
        end = PREDICATES_OFFSET + num_children + ( \
            PREDICATE_INFO_LENGTH * num_predicates)
        children = node[slice(start, end)]
        for child in children:
            if apply_predicates(child, input_data, fields):
                return calculate_depth(child, input_data, fields, depth)

    return depth


def use_cache(cache_get):
    """Checks whether the user has provided a cache get function to retrieve
       local models.

    """
    return cache_get is not None and hasattr(cache_get, '__call__')


class Anomaly(ModelFields):
    """ A minimal anomaly detector designed to build quickly from a
    specialized external representation. See file documentation, above,
    for usage.

    """

    def __init__(self, anomaly, api=None, cache_get=None):

        if use_cache(cache_get):
            # using a cache to store the Minomaly attributes
            anomaly_id = get_anomaly_id(anomaly)
            self.__dict__ = msgpack.loads(cache_get(anomaly_id))
        else:
            self.resource_id = None
            self.sample_size = None
            self.input_fields = None
            self.mean_depth = None
            self.expected_mean_depth = None
            self.iforest = None
            self.id_fields = []
            api = get_api_connection(api)
            self.resource_id, anomaly = get_resource_dict(
                anomaly, "anomaly", api=api)

            if 'object' in anomaly and isinstance(anomaly['object'], dict):
                anomaly = anomaly['object']
                self.sample_size = anomaly.get('sample_size')
                self.input_fields = anomaly.get('input_fields')
                self.id_fields = anomaly.get('id_fields', [])

            if 'model' in anomaly and isinstance(anomaly['model'], dict):
                ModelFields.__init__(
                    self, anomaly['model'].get('fields'),
                    missing_tokens=anomaly['model'].get('missing_tokens'))

                self.mean_depth = anomaly['model'].get('mean_depth')
                self.normalization_factor = anomaly['model'].get(
                    'normalization_factor')
                self.nodes_mean_depth = anomaly['model'].get(
                    'nodes_mean_depth')
                status = get_status(anomaly)
                if 'code' in status and status['code'] == FINISHED:
                    self.expected_mean_depth = None
                    if self.mean_depth is None or self.sample_size is None:
                        raise Exception("The anomaly data is not complete. "
                                        "Score will not be available")
                    else:
                        self.norm = self.normalization_factor if \
                            self.normalization_factor is not None else \
                            self.norm_factor()
                    iforest = anomaly['model'].get('trees', [])
                    if iforest:
                        self.iforest = [
                            build_tree([anomaly_tree['root']])
                            for anomaly_tree in iforest]
                    self.top_anomalies = anomaly['model']['top_anomalies']
                else:
                    raise Exception("The anomaly isn't finished yet")

    def norm_factor(self):
        """Computing the normalization factor for simple anomaly detectors"""
        if self.mean_depth is not None:
            default_depth = self.mean_depth if self.sample_size == 1 else \
                (2 * (DEPTH_FACTOR + math.log(self.sample_size - 1) -
                      (float(self.sample_size - 1) / self.sample_size)))
            return min(self.mean_depth, default_depth)

    def anomaly_score(self, input_data):
        """Returns the anomaly score given by the iforest
            To produce an anomaly score, we evaluate each tree in the iforest
            for its depth result (see the depth method in the AnomalyTree
            object for details). We find the average of these depths
            to produce an `observed_mean_depth`. We calculate an
            `expected_mean_depth` using the `sample_size` and `mean_depth`
            parameters which come as part of the forest message.
            We combine those values as seen below, which should result in a
            value between 0 and 1.

        """
        # corner case with only one record
        if self.sample_size == 1 and self.normalization_factor is None:
            return 1
        # Checks and cleans input_data leaving the fields used in the model
        input_data = self.filter_input_data(input_data)

        # Strips affixes for numeric values and casts to the final field type
        cast(input_data, self.fields)

        depth_sum = 0

        if self.iforest is None:
            raise Exception("We could not find the iforest information to "
                            "compute the anomaly score. Please, rebuild your "
                            "Anomaly object from a complete anomaly detector "
                            "resource.")
        for tree in self.iforest:
            tree_depth = calculate_depth(tree[1], input_data, self.fields)
            depth_sum += tree_depth

        observed_mean_depth = float(depth_sum) / len(self.iforest)
        return math.pow(2, - observed_mean_depth / self.norm)

    def anomalies_filter(self, include=True):
        """Returns the LISP expression needed to filter the subset of
           top anomalies. When include is set to True, only the top
           anomalies are selected by the filter. If set to False, only the
           rest of the dataset is selected.
        """
        anomaly_filters = []
        for anomaly in self.top_anomalies:
            filter_rules = []
            row = anomaly.get('row', [])
            for index, value in enumerate(row):
                field_id = self.input_fields[index]
                if field_id in self.id_fields:
                    continue
                if value is None or value is "":
                    filter_rules.append('(missing? "%s")' % field_id)
                else:
                    if (self.fields[field_id]["optype"]
                            in ["categorical", "text"]):
                        value = json.dumps(value)
                    filter_rules.append('(= (f "%s") %s)' % (field_id, value))
            if filter_rules:
                anomaly_filters.append("(and %s)" % " ".join(filter_rules))

        anomalies_filter = " ".join(anomaly_filters)
        if include:
            if len(anomaly_filters) == 1:
                return anomalies_filter
            return "(or %s)" % anomalies_filter
        else:
            return "(not (or %s))" % anomalies_filter

    def dump(self, output=None, cache_set=None):
        """Uses msgpack to serialize the anomaly object
        If cache_set is filled with a cache set method, the method is called

        """
        if use_cache(cache_set):
            dump_string= msgpack.dumps(self.__dict__)
            cache_set(self.resource_id, dump_string)
        else:
            msgpack.pack(self.__dict__, output)

    def dumps(self, cache_set=None):
        """Uses msgpack to serialize the anomaly object to a string

        """
        return msgpack.dumps(self.__dict__)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base class for whizzml script' REST calls

   https://bigml.com/api/scripts

"""

import os
import requests
import re
try:
    import simplejson as json
except ImportError:
    import json

from urllib.parse import urljoin

from bigml.api_handlers.resourcehandler import ResourceHandler
from bigml.api_handlers.resourcehandler import check_resource_type, \
    get_script_id, get_resource_type, check_resource
from bigml.constants import SCRIPT_PATH, TINY_RESOURCE
from bigml.util import is_url
from bigml.bigmlconnection import HTTP_OK


def retrieve_script_args(gist_url):
    """Retrieves the information to create a script from a public
       gist url

    """

    response = requests.get(gist_url)
    response.encoding = "utf8"
    if response.status_code == HTTP_OK:
        pattern = r"\"[^\"]*?\/raw\/[^\"]*"
        urls = re.findall(pattern, response.text)
        script_args = {}

        for url in urls:
            url = urljoin(gist_url, url.replace("\"", ""))
            if url.endswith(".whizzml"):
                response = requests.get(url)
                if response.status_code == HTTP_OK:
                    script_args["source_code"] = response.text
            if url.endswith(".json"):
                response = requests.get(url, \
                    headers={"content-type": "application/json"})
                if response.status_code == HTTP_OK:
                    script_args["json"] = response.text
        return script_args
    else:
        raise ValueError("The url did not contain the expected structure.")


class ScriptHandler(ResourceHandler):
    """This class is used by the BigML class as
       a mixin that provides the whizzml script' REST calls. It should not
       be instantiated independently.

    """
    def __init__(self):
        """Initializes the ScriptHandler. This class is intended to be
           used as a mixin on ResourceHandler, that inherits its
           attributes and basic method from BigMLConnection, and must not be
           instantiated independently.

        """
        self.script_url = self.url + SCRIPT_PATH

    def create_script(self, source_code=None, args=None,
                      wait_time=3, retries=10):
        """Creates a whizzml script from its source code. The `source_code`
           parameter can be a:
            {script ID}: the ID for an existing whizzml script
            {path}: the path to a file containing the source code
            {string} : the string containing the source code for the script

        """
        create_args = {}
        if args is not None:
            create_args.update(args)

        if source_code is None:
            raise Exception('A valid code string'
                            ' or a script id must be provided.')
        resource_type = get_resource_type(source_code)
        if resource_type == SCRIPT_PATH:
            script_id = get_script_id(source_code)
            if script_id:
                check_resource(script_id,
                               query_string=TINY_RESOURCE,
                               wait_time=wait_time, retries=retries,
                               raise_on_error=True, api=self)
                create_args.update({
                    "origin": script_id})
        elif isinstance(source_code, str):
            if is_url(source_code):
                script_args = retrieve_script_args(source_code)
                source_code = script_args.get("source_code")
                create_args.update(json.loads(script_args.get("json")))
            else:
                try:
                    if os.path.exists(source_code):
                        with open(source_code) as code_file:
                            source_code = code_file.read()
                except IOError:
                    raise IOError("Could not open the source code file %s." %
                                  source_code)
            create_args.update({
                "source_code": source_code})
        else:
            raise Exception("A script id or a valid source code"
                            " is needed to create a"
                            " script. %s found." % resource_type)


        body = json.dumps(create_args)
        return self._create(self.script_url, body)

    def get_script(self, script, query_string=''):
        """Retrieves a script.

           The script parameter should be a string containing the
           script id or the dict returned by create_script.
           As script is an evolving object that is processed
           until it reaches the FINISHED or FAULTY state, the function will
           return a dict that encloses the script content and state info
           available at the time it is called.
        """
        check_resource_type(script, SCRIPT_PATH,
                            message="A script id is needed.")
        script_id = get_script_id(script)
        if script_id:
            return self._get("%s%s" % (self.url, script_id),
                             query_string=query_string)

    def list_scripts(self, query_string=''):
        """Lists all your scripts.

        """
        return self._list(self.script_url, query_string)

    def update_script(self, script, changes):
        """Updates a script.

        """
        check_resource_type(script, SCRIPT_PATH,
                            message="A script id is needed.")
        script_id = get_script_id(script)
        if script_id:
            body = json.dumps(changes)
            return self._update("%s%s" % (self.url, script_id), body)

    def delete_script(self, script):
        """Deletes a script.

        """
        check_resource_type(script, SCRIPT_PATH,
                            message="A script id is needed.")
        script_id = get_script_id(script)
        if script_id:
            return self._delete("%s%s" % (self.url, script_id))



# -*- coding: utf-8 -*-
#
# Copyright 2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Common auxiliar functions to be used in the node predicate evaluation
"""
import operator
import re

# Operator Codes
LT = 0
LE = 1
EQ = 2
NE = 3
GE = 4
GT = 5
IN = 6

# Map operator string to its corresponding code
OPERATOR_CODE = {"<": LT,
                 "<=": LE,
                 "=": EQ,
                 "!=": NE,
                 "/=": NE,
                 ">=": GE,
                 ">": GT,
                 "in": IN}

# Map operator code to its corresponding function
OPERATOR = [operator.lt,
            operator.le,
            operator.eq,
            operator.ne,
            operator.ge,
            operator.gt,
            operator.contains]

TM_TOKENS = 'tokens_only'
TM_FULL_TERM = 'full_terms_only'
TM_ALL = 'all'
FULL_TERM_PATTERN = re.compile(r'^.+\b.+$', re.U)

OPERATION_OFFSET = 2
FIELD_OFFSET = 3
VALUE_OFFSET = 4
TERM_OFFSET = 5
MISSING_OFFSET = 6

PREDICATE_INFO_LENGTH = 5

def term_matches(text, forms_list, options):
    """ Counts the number of occurences of the words in forms_list in the text
    The terms in forms_list can either be tokens or full terms. The
    matching for tokens is contains and for full terms is equals.
    """
    token_mode = options.get('token_mode', TM_TOKENS)
    case_sensitive = options.get('case_sensitive', False)
    first_term = forms_list[0]
    if token_mode == TM_FULL_TERM:
        return full_term_match(text, first_term, case_sensitive)

    # In token_mode='all' we will match full terms using equals and
    # tokens using contains
    if token_mode == TM_ALL and len(forms_list) == 1:
        if re.match(FULL_TERM_PATTERN, first_term):
            return full_term_match(text, first_term, case_sensitive)

    return term_matches_tokens(text, forms_list, case_sensitive)


def full_term_match(text, full_term, case_sensitive):
    """Counts the match for full terms according to the case_sensitive option
    """
    if not case_sensitive:
        text = text.lower()
        full_term = full_term.lower()

    return 1 if text == full_term else 0


def get_tokens_flags(case_sensitive):
    """Returns flags for regular expression matching depending on text analysis
    options
    """
    flags = re.U
    if not case_sensitive:
        flags = (re.I | flags)

    return flags


def term_matches_tokens(text, forms_list, case_sensitive):
    """Counts the number of occurences of the words in forms_list in the text
    """
    flags = get_tokens_flags(case_sensitive)
    expression = r'(\b|_)%s(\b|_)' % '(\\b|_)|(\\b|_)'.join(forms_list)
    pattern = re.compile(expression, flags=flags)
    matches = re.findall(pattern, text)
    return len(matches)


def item_matches(text, item, options):
    """ Counts the number of occurences of the item in the text
    The matching considers the separator or
    the separating regular expression.
    """
    separator = options.get('separator', ' ')
    regexp = options.get('separator_regexp')
    if regexp is None:
        regexp = r"%s" % re.escape(separator)

    return count_items_matches(text, item, regexp)


def count_items_matches(text, item, regexp):
    """ Counts the number of occurences of the item in the text
    """
    expression = r'(^|%s)%s($|%s)' % (regexp, item, regexp)
    pattern = re.compile(expression, flags=re.U)
    matches = re.findall(pattern, text)

    return len(matches)

def apply_predicates(node, input_data, fields):
    num_predicates = node[1]

    for i in range(num_predicates):
        operation = node[OPERATION_OFFSET + (PREDICATE_INFO_LENGTH * i)]
        field = node[FIELD_OFFSET + (PREDICATE_INFO_LENGTH * i)]
        value = node[VALUE_OFFSET + (PREDICATE_INFO_LENGTH * i)]
        term = node[TERM_OFFSET + (PREDICATE_INFO_LENGTH * i)]
        missing = node[MISSING_OFFSET + (PREDICATE_INFO_LENGTH * i)]

        if not apply_predicate(operation, field, value, term, missing,
                               input_data, fields[field]):
            return False

    return True


def apply_predicate(operator, field, value, term, missing, input_data,
                    field_info):
    """Applies the operators defined in the predicate as strings to
    the provided input data
    """
    # for missing operators
    if input_data.get(field) is None:
        # text and item fields will treat missing values by following the
        # doesn't contain branch
        if term is None:
            return missing or (
                operator == EQ and value is None)
    elif operator == NE and value is None:
        return True

    if term is not None:
        if field_info['optype'] == 'text':
            all_forms = field_info['summary'].get('term_forms', {})
            term_forms = all_forms.get(term, [])
            terms = [term]
            terms.extend(term_forms)
            options = field_info['term_analysis']
            input_terms = term_matches(input_data.get(field, ""), terms,
                                       options)
            return OPERATOR[operator](input_terms, value)
        else:
            # new items optype
            options = field_info['item_analysis']
            input_items = item_matches(input_data.get(field, ""), term,
                                       options)
            return OPERATOR[operator](input_items, value)
    if operator == IN:
        return OPERATOR[operator](value, input_data[field])
    return OPERATOR[operator](input_data[field], value)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating correlation

"""
from .world import world, setup_module, teardown_module
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_correlation_steps as correlation_create

class TestCorrelation(object):

    def setup(self):
        """
            Debug information
        """
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)

    def test_scenario1(self):
        """
            Scenario: Successfully creating a correlation from a dataset:
                Given I create a data source uploading a "<data>" file
                And I wait until the source is ready less than <time_1> secs
                And I create a dataset
                And I wait until the dataset is ready less than <time_2> secs
                And I create a correlation from a dataset
                And I wait until the correlation is ready less than <time_3> secs
                And I update the correlation name to "<correlation_name>"
                When I wait until the correlation is ready less than <time_4> secs
                Then the correlation name is "<correlation_name>"

                Examples:
                | data                | time_1  | time_2 | time_3 | time_4 | correlation_name |
                | ../data/iris.csv | 10      | 10     | 20     | 20 | my new correlation name |
        """
        print(self.test_scenario1.__doc__)
        examples = [
            ['data/iris.csv', '10', '10', '20', '20', 'my new correlation name']]
        for example in examples:
            print("\nTesting with:\n", example)
            source_create.i_upload_a_file(self, example[0])
            source_create.the_source_is_finished(self, example[1])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(self, example[2])
            correlation_create.i_create_a_correlation_from_dataset(self)
            correlation_create.the_correlation_is_finished_in_less_than(self, example[3])
            correlation_create.i_update_correlation_name(self, example[5])
            correlation_create.the_correlation_is_finished_in_less_than(self, example[4])
            correlation_create.i_check_correlation_name(self, example[5])



# -*- coding: utf-8 -*-
#
# Copyright 2014-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base class for batch anomaly scores' REST calls

   https://bigml.com/api/batchanomalyscores

"""

try:
    import simplejson as json
except ImportError:
    import json

from bigml.bigmlconnection import DOWNLOAD_DIR
from bigml.api_handlers.resourcehandler import ResourceHandler
from bigml.api_handlers.resourcehandler import check_resource_type, \
    get_batch_anomaly_score_id
from bigml.constants import BATCH_ANOMALY_SCORE_PATH, ANOMALY_PATH


class BatchAnomalyScoreHandler(ResourceHandler):
    """This class is used by the BigML class as
       a mixin that provides the REST calls models. It should not
       be instantiated independently.

    """
    def __init__(self):
        """Initializes the BatchAnomalyScoreHandler. This class is intended
           to be used as a mixin on ResourceHandler, that inherits its
           attributes and basic method from BigMLConnection, and must not be
           instantiated independently.

        """
        self.batch_anomaly_score_url = self.prediction_base_url + \
            BATCH_ANOMALY_SCORE_PATH

    def create_batch_anomaly_score(self, anomaly, dataset,
                                   args=None, wait_time=3, retries=10):
        """Creates a new batch anomaly score.


        """
        create_args = {}
        if args is not None:
            create_args.update(args)

        origin_resources_checked = self.check_origins(
            dataset, anomaly, create_args, model_types=[ANOMALY_PATH],
            wait_time=wait_time, retries=retries)

        if origin_resources_checked:
            body = json.dumps(create_args)
            return self._create(self.batch_anomaly_score_url, body)

    def get_batch_anomaly_score(self, batch_anomaly_score, query_string=''):
        """Retrieves a batch anomaly score.

           The batch_anomaly_score parameter should be a string containing the
           batch_anomaly_score id or the dict returned by
           create_batch_anomaly_score.
           As batch_anomaly_score is an evolving object that is processed
           until it reaches the FINISHED or FAULTY state, the function will
           return a dict that encloses the batch_anomaly_score values and state
           info available at the time it is called.
        """
        check_resource_type(batch_anomaly_score, BATCH_ANOMALY_SCORE_PATH,
                            message="A batch anomaly score id is needed.")
        batch_anomaly_score_id = get_batch_anomaly_score_id(
            batch_anomaly_score)
        if batch_anomaly_score_id:
            return self._get("%s%s" % (self.url, batch_anomaly_score_id),
                             query_string=query_string)

    def download_batch_anomaly_score(self, batch_anomaly_score, filename=None):
        """Retrieves the batch anomaly score file.

           Downloads anomaly scores, that are stored in a remote CSV file. If
           a path is given in filename, the contents of the file are downloaded
           and saved locally. A file-like object is returned otherwise.
        """
        check_resource_type(batch_anomaly_score, BATCH_ANOMALY_SCORE_PATH,
                            message="A batch anomaly score id is needed.")
        batch_anomaly_score_id = get_batch_anomaly_score_id(
            batch_anomaly_score)
        if batch_anomaly_score_id:
            return self._download("%s%s%s" % (self.url, batch_anomaly_score_id,
                                              DOWNLOAD_DIR), filename=filename)

    def list_batch_anomaly_scores(self, query_string=''):
        """Lists all your batch anomaly scores.

        """
        return self._list(self.batch_anomaly_score_url, query_string)

    def update_batch_anomaly_score(self, batch_anomaly_score, changes):
        """Updates a batch anomaly scores.

        """
        check_resource_type(batch_anomaly_score, BATCH_ANOMALY_SCORE_PATH,
                            message="A batch anomaly score id is needed.")
        batch_anomaly_score_id = get_batch_anomaly_score_id(
            batch_anomaly_score)
        if batch_anomaly_score_id:
            body = json.dumps(changes)
            return self._update("%s%s" % (self.url,
                                          batch_anomaly_score_id), body)

    def delete_batch_anomaly_score(self, batch_anomaly_score):
        """Deletes a batch anomaly score.

        """
        check_resource_type(batch_anomaly_score, BATCH_ANOMALY_SCORE_PATH,
                            message="A batch anomaly score id is needed.")
        batch_anomaly_score_id = get_batch_anomaly_score_id(
            batch_anomaly_score)
        if batch_anomaly_score_id:
            return self._delete("%s%s" % (self.url, batch_anomaly_score_id))



# -*- coding: utf-8 -*-
#
# Copyright 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Downloading dataset

"""
from .world import world, setup_module, teardown_module
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_model_steps as model_create


class TestDownload(object):

    def setup(self):
        """
            Debug information
        """
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)

    def test_scenario1(self):
        """

            Scenario: Successfully exporting a dataset:
                Given I create a data source uploading a "<data>" file
                And I wait until the source is ready less than <time_1> secs
                And I create a dataset
                And I wait until the dataset is ready less than <time_2> secs
                And I download the dataset file to "<local_file>"
                Then file "<local_file>" is like file "<data>"

                Examples:
                | data             | time_1  | time_2 | local_file |
                | ../data/iris.csv | 30      | 30     | ./tmp/exported_iris.csv |
        """
        print(self.test_scenario1.__doc__)
        examples = [
            ['data/iris.csv', '30', '30', 'tmp/exported_iris.csv']]
        for example in examples:
            print("\nTesting with:\n", example)
            source_create.i_upload_a_file(self, example[0])
            source_create.the_source_is_finished(self, example[1])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(self, example[2])
            dataset_create.i_export_a_dataset(self, example[3])
            dataset_create.files_equal(self, example[3], example[0])

    def test_scenario2(self):
        """
            Scenario: Successfully creating a model and exporting it:
                Given I create a data source uploading a "<data>" file
                And I wait until the source is ready less than <time_1> secs
                And I create a dataset
                And I wait until the dataset is ready less than <time_2> secs
                And I create a model
                And I wait until the model is ready less than <time_3> secs
                And I export the <"pmml"> model to file "<expected_file>"
                Then I check the model is stored in "<expected_file>" file in <"pmml">

                Examples:
                | data                   | time_1  | time_2 | time_3 | expected_file         | pmml
                | data/iris.csv          | 10      | 10     | 10     | tmp/model/iris.json   | false
                | data/iris_sp_chars.csv | 10      | 10     | 10     | tmp/model/iris_sp_chars.pmml   | true

        """
        print(self.test_scenario2.__doc__)
        examples = [
            ['data/iris.csv', '30', '30', '30', 'tmp/model/iris.json', False],
            ['data/iris_sp_chars.csv', '30', '30', '30', 'tmp/model/iris_sp_chars.pmml', True]]
        for example in examples:
            print("\nTesting with:\n", example)
            source_create.i_upload_a_file(self, example[0])
            source_create.the_source_is_finished(self, example[1])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(self, example[2])
            model_create.i_create_a_model(self)
            model_create.the_model_is_finished_in_less_than(self, example[3])
            model_create.i_export_model(self, example[5], example[4])
            model_create.i_check_model_stored(self, example[4], example[5])



# -*- coding: utf-8 -*-
#
# Copyright 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating and updating scripts

"""
from .world import world, setup_module, teardown_module
from . import create_script_steps as script_create
from . import create_execution_steps as execution_create

class TestExecution(object):

    def setup(self):
        """
            Debug information
        """
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)

    def test_scenario1(self):
        """
            Scenario: Successfully creating a whizzml script execution:
                Given I create a whizzml script from a excerpt of code "<source_code>"
                And I wait until the script is ready less than <time_1> secs
                And I create a whizzml script execution from an existing script
                And I wait until the execution is ready less than <time_2> secs
                And I update the execution with "<param>", "<param_value>"
                And I wait until the execution is ready less than <time_3> secs
                And I create a local execution
                Then the script id is correct, the value of "<param>" is "<param_value>" and the result is "<result>"
                And the local execution result is "<result>"

                Examples:

                | source_code      | time_1  | time_2  | time_3  | param | param_value | result
                | (+ 1 1)          | 10      | 10      | 10      | name  | my execution | 2
        """
        print(self.test_scenario1.__doc__)
        examples = [
            ['(+ 1 1)', '10', '10', '10', 'name', 'my execution', 2]]

        for example in examples:
            print("\nTesting with:\n", example)
            script_create.i_create_a_script(self, example[0])
            script_create.the_script_is_finished(self, example[1])
            execution_create.i_create_an_execution(self)
            execution_create.the_execution_is_finished(self, example[2])
            execution_create.i_update_an_execution(self, example[4], example[5])
            execution_create.the_execution_is_finished(self, example[3])
            execution_create.create_local_execution(self)
            execution_create.the_execution_and_attributes(self, example[4], example[5], example[6])
            execution_create.the_local_execution_result_is(self, example[6])

    def test_scenario2(self):
        """
            Scenario: Successfully creating a whizzml script execution from a list of scripts:
                Given I create a whizzml script from a excerpt of code "<source_code>"
                And I wait until the script is ready less than <time_1> secs
                And I create a whizzml script from a excerpt of code "<source_code>"
                And I wait until the script is ready less than <time_1> secs
                And I create a whizzml script execution from the last two scripts
                And I wait until the execution is ready less than <time_2> secs
                And I update the execution with "<param>", "<param_value>"
                And I wait until the execution is ready less than <time_3> secs
                Then the script ids are correct, the value of "<param>" is "<param_value>" and the result is "<result>"

                Examples:

                | source_code      | time_1  | time_2  | time_3  | param | param_value | result
                | (+ 1 1)          | 10      | 10      | 10      | name  | my execution | [2, 2]
        """
        print(self.test_scenario2.__doc__)
        examples = [
            ['(+ 1 1)', '10', '10', '10', 'name', 'my execution', [2, 2]]]
        for example in examples:
            print("\nTesting with:\n", example)
            script_create.i_create_a_script(self, example[0])
            script_create.the_script_is_finished(self, example[1])
            script_create.i_create_a_script(self, example[0])
            script_create.the_script_is_finished(self, example[1])
            execution_create.i_create_an_execution_from_list(self, 2)
            execution_create.the_execution_is_finished(self, example[2])
            execution_create.i_update_an_execution(self, example[4], example[5])
            execution_create.the_execution_is_finished(self, example[3])
            execution_create.the_execution_ids_and_attributes(self, 2, example[4], example[5], example[6])

    def test_scenario3(self):
        """
            Scenario: Successfully creating a whizzml script execution from a local or remote file:
                Given I create a whizzml script from a excerpt of code "<source_code>"
                And I wait until the script is ready less than <time_1> secs
                And I create a whizzml script from a excerpt of code "<source_code>"
                And I wait until the script is ready less than <time_1> secs
                And I create a whizzml script execution from the last two scripts
                And I wait until the execution is ready less than <time_2> secs
                And I update the execution with "<param>", "<param_value>"
                And I wait until the execution is ready less than <time_3> secs
                Then the script ids are correct, the value of "<param>" is "<param_value>" and the result is "<result>"

                Examples:

                | source_code      | time_1  | time_2  | time_3  | param | param_value | result
        """
        print(self.test_scenario3.__doc__)
        examples = [
            ['data/one_plus_one.whizzml', '10', '10', '10', 'name', 'my execution', 2],
            ['https://gist.github.com/mmerce/49e0a69cab117b6a11fb490140326020', '10', '10', '10', 'name', 'my execution', 2]

]
        for example in examples:
            print("\nTesting with:\n", example)
            script_create.i_create_a_script_from_file_or_url(self, example[0])
            script_create.the_script_is_finished(self, example[1])
            execution_create.i_create_an_execution(self)
            execution_create.the_execution_is_finished(self, example[2])
            execution_create.i_update_an_execution(self, example[4], example[5])
            execution_create.the_execution_is_finished(self, example[3])
            execution_create.the_execution_and_attributes(self, example[4], example[5], example[6])



# -*- coding: utf-8 -*-
#
# Copyright 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base class for statisticaltests' REST calls

   https://bigml.com/api/statisticaltests

"""

try:
    import simplejson as json
except ImportError:
    import json


from bigml.api_handlers.resourcehandler import ResourceHandler
from bigml.api_handlers.resourcehandler import check_resource_type, \
    get_statistical_test_id, get_resource_type, get_dataset_id, check_resource
from bigml.constants import (STATISTICAL_TEST_PATH, DATASET_PATH,
                             TINY_RESOURCE)


class StatisticalTestHandler(ResourceHandler):
    """This class is used by the BigML class as
       a mixin that provides the statistical tests' REST calls. It should not
       be instantiated independently.

    """
    def __init__(self):
        """Initializes the StatisticalTestHandler. This class is intended to be
           used as a mixin on ResourceHandler, that inherits its
           attributes and basic method from BigMLConnection, and must not be
           instantiated independently.

        """
        self.statistical_test_url = self.url + STATISTICAL_TEST_PATH

    def create_statistical_test(self, dataset, args=None, wait_time=3, retries=10):
        """Creates a statistical test from a `dataset`.

        """
        dataset_id = None
        resource_type = get_resource_type(dataset)
        if resource_type == DATASET_PATH:
            dataset_id = get_dataset_id(dataset)
            check_resource(dataset_id,
                           query_string=TINY_RESOURCE,
                           wait_time=wait_time, retries=retries,
                           raise_on_error=True, api=self)
        else:
            raise Exception("A dataset id is needed to create a"
                            " statistical test. %s found." % resource_type)

        create_args = {}
        if args is not None:
            create_args.update(args)
        create_args.update({
            "dataset": dataset_id})

        body = json.dumps(create_args)
        return self._create(self.statistical_test_url, body)

    def get_statistical_test(self, statistical_test, query_string=''):
        """Retrieves a statistical test.

           The statistical test parameter should be a string containing the
           statisticaltest id or the dict returned by create_statistical_test.
           As an statistical test is an evolving object that is processed
           until it reaches the FINISHED or FAULTY state, the function will
           return a dict that encloses the statistical test values and state
           info available at the time it is called.
        """
        check_resource_type(statistical_test, STATISTICAL_TEST_PATH,
                            message="A statistical test id is needed.")
        statistical_test_id = get_statistical_test_id(statistical_test)
        if statistical_test_id:
            return self._get("%s%s" % (self.url, statistical_test_id),
                             query_string=query_string)

    def list_statistical_tests(self, query_string=''):
        """Lists all your statistical tests.

        """
        return self._list(self.statistical_test_url, query_string)

    def update_statistical_test(self, statistical_test, changes):
        """Updates an statistical test.

        """
        check_resource_type(statistical_test, STATISTICAL_TEST_PATH,
                            message="A statistical test id is needed.")
        statistical_test_id = get_statistical_test_id(statistical_test)
        if statistical_test_id:
            body = json.dumps(changes)
            return self._update("%s%s" % (self.url, statistical_test_id), body)

    def delete_statistical_test(self, statistical_test):
        """Deletes a statistical test.

        """
        check_resource_type(statistical_test, STATISTICAL_TEST_PATH,
                            message="A statistical test id is needed.")
        statistical_test_id = get_statistical_test_id(statistical_test)
        if statistical_test_id:
            return self._delete("%s%s" % (self.url, statistical_test_id))



# -*- coding: utf-8 -*-

#
# Copyright 2014-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import time
import json
from datetime import datetime
from urllib.parse import urlencode
from nose.tools import eq_, assert_less
from .world import world, logged_wait

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED
from bigml.api import FAULTY
from bigml.api import UPLOADING
from bigml.api import get_status

from .read_project_steps import i_get_the_project


def i_create_project(step, name):
    resource = world.api.create_project({"name": name})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.project = resource['object']
    # save reference
    world.projects.append(resource['resource'])


def wait_until_project_status_code_is(step, code1, code2, secs):
    start = datetime.utcnow()
    delta = int(secs) * world.delta
    i_get_the_project(step, world.project['resource'])
    status = get_status(world.project)
    count = 0
    while (status['code'] != int(code1) and
           status['code'] != int(code2)):
        count += 1
        logged_wait(start, delta, count, "project")
        assert_less((datetime.utcnow() - start).seconds, delta)
        i_get_the_project(step, world.project['resource'])
        status = get_status(world.project)
    eq_(status['code'], int(code1))


def the_project_is_finished(step, secs):
    wait_until_project_status_code_is(step, FINISHED, FAULTY, secs)


def i_update_project_name_with(step, name=""):
    resource = world.api.update_project(world.project.get('resource'),
                                        {"name": name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.project = resource['object']


def i_check_project_name(step, name=""):
    updated_name = world.project.get("name", "")
    eq_(updated_name, name)



# -*- coding: utf-8 -*-
#
# Copyright 2012, 2015-2020 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import json
from datetime import datetime
from .world import world, logged_wait
from nose.tools import eq_, assert_less, assert_greater

from bigml.api import HTTP_CREATED
from bigml.api import FINISHED
from bigml.api import FAULTY
from bigml.api import get_status

from .read_evaluation_steps import i_get_the_evaluation

#@step(r'I create an evaluation for the model with the dataset$')
def i_create_an_evaluation(step):
    dataset = world.dataset.get('resource')
    model = world.model.get('resource')
    resource = world.api.create_evaluation(model, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


#@step(r'I create an evaluation for the ensemble with the dataset$')
def i_create_an_evaluation_ensemble(step, params=None):
    if params is None:
        params = {}
    dataset = world.dataset.get('resource')
    ensemble = world.ensemble.get('resource')
    resource = world.api.create_evaluation(ensemble, dataset, params)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])

#@step(r'I create an evaluation for the logistic regression with the dataset$')
def i_create_an_evaluation_logistic(step):
    dataset = world.dataset.get('resource')
    logistic = world.logistic_regression.get('resource')
    resource = world.api.create_evaluation(logistic, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])

#@step(r'I create an evaluation for the deepnet with the dataset$')
def i_create_an_evaluation_deepnet(step):
    dataset = world.dataset.get('resource')
    deepnet = world.deepnet.get('resource')
    resource = world.api.create_evaluation(deepnet, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


#@step(r'I create an evaluation for the fusion with the dataset$')
def i_create_an_evaluation_fusion(step):
    dataset = world.dataset.get('resource')
    fusion = world.fusion.get('resource')
    resource = world.api.create_evaluation(fusion, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])

#@step(r'I wait until the evaluation status code is either (\d) or (-\d) less than (\d+)')
def wait_until_evaluation_status_code_is(step, code1, code2, secs):
    start = datetime.utcnow()
    delta = int(secs) * world.delta
    i_get_the_evaluation(step, world.evaluation['resource'])
    status = get_status(world.evaluation)
    count = 0
    while (status['code'] != int(code1) and
           status['code'] != int(code2)):
        count += 1
        logged_wait(start, delta, count, "evaluation")
        assert_less((datetime.utcnow() - start).seconds, delta)
        i_get_the_evaluation(step, world.evaluation['resource'])
        status = get_status(world.evaluation)
    eq_(status['code'], int(code1))

#@step(r'I wait until the evaluation is ready less than (\d+)')
def the_evaluation_is_finished_in_less_than(step, secs):
    wait_until_evaluation_status_code_is(step, FINISHED, FAULTY, secs)

#@step(r'the measured "(.*)" is (\d+\.*\d*)')
def the_measured_measure_is_value(step, measure, value):
    ev = world.evaluation['result']['model'][measure] + 0.0
    eq_(ev, float(value), "The %s is: %s and %s is expected" % (
        measure, ev, float(value)))

#@step(r'the measured "(.*)" is greater than (\d+\.*\d*)')
def the_measured_measure_is_greater_value(step, measure, value):
    assert_greater(world.evaluation['result']['model'][measure] + 0.0,
                   float(value))
